function [Z,UtZ,Zchol] = squareLoss_dual_get_Z(U, lambda, I, J, Y, mask)
    Ut = U.';
    numThreads = 4; % parallelizing some components of the inner problem over tasks. Number should be increased if system supports it. 
    parallel_flag = false;% default value should be true. It is advised to contact authors if one is not able to run the code with parallel_flag as true. 
    UomegatYomega = mymultfullsparse(Ut, Y, mask); % |Omega| operations. Mex code heavy.
    if parallel_flag
        Zchol = myspbuildmatrix_parallel(mask, Ut, lambda, numThreads);
        UtUinvUtY = blocksolve_parallel(Zchol, UomegatYomega, numThreads);
        UUtUinvUtY = myspmaskmult_parallel(Ut, UtUinvUtY, I, J, numThreads); % |Omega| operations. Mex code heavy.
    else
        Zchol = myspbuildmatrix(mask, Ut, lambda);
        UtUinvUtY = blocksolve(Zchol, UomegatYomega);
        UUtUinvUtY = myspmaskmult(Ut, UtUinvUtY, I, J); % |Omega| operations. Mex code heavy.
    end
    Z = (2*lambda)*(Y - UUtUinvUtY);
    UtZ = mymultfullsparse(Ut, Z, mask); % |Omega| operations. Mex code heavy.
    
end

function X = blocksolve_parallel(R, B, numThreads)
    % R is a cell containing n upper triangular square matrices of size r.
    % B is a matrix of size r x n. X is a matrix of size r x n where the i-th
    % column is the solution of the equation A{i}x = B(:, i), with
    % A{i} = R{i}.'*R{i}. This is: the R{i} matrices are the upper triangular
    % Cholesky factors of the matrices A{i} (which we do not require/store).
    
    X = mycholsolvecell_parallel(R, B, numThreads);    
end

function X = blocksolve(R, B)
    % R is a cell containing n upper triangular square matrices of size r.
    % B is a matrix of size r x n. X is a matrix of size r x n where the i-th
    % column is the solution of the equatio`n A{i}x = B(:, i), with
    % A{i} = R{i}.'*R{i}. This is: the R{i} matrices are the upper triangular
    % Cholesky factors of the matrices A{i} (which we do not require/store).
    
    X = mycholsolvecell(R,B);
end
